'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.ArrayZipper = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

exports.makeZipper = makeZipper;

var _assoc = require('ramda/src/assoc');

var _assoc2 = _interopRequireDefault(_assoc);

var _head = require('ramda/src/head');

var _head2 = _interopRequireDefault(_head);

var _init = require('ramda/src/init');

var _init2 = _interopRequireDefault(_init);

var _last = require('ramda/src/last');

var _last2 = _interopRequireDefault(_last);

var _tail = require('ramda/src/tail');

var _tail2 = _interopRequireDefault(_tail);

var _pipe = require('ramda/src/pipe');

var _pipe2 = _interopRequireDefault(_pipe);

var _merge = require('ramda/src/merge');

var _merge2 = _interopRequireDefault(_merge);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var TOP = null;
var END = '@@zipper/END';

function isEmpty(arr) {
    return !arr.length;
}

function whilst(predicate, fn, value) {
    var curr = value;
    while (predicate(curr)) {
        curr = fn(curr);
    }
    return curr;
}

function makeZipper(_isBranch, _getChildren, makeItem) {

    var getItem = _head2.default;
    var getPath = _last2.default;

    var isBranch = (0, _pipe2.default)(getItem, _isBranch);

    function isLeaf(loc) {
        return !isBranch(loc);
    }

    function _isEnd(loc) {
        return getPath(loc) === END;
    }

    function getParentItems(loc) {
        return getPath(loc).parentItems;
    }

    function isTop(loc) {
        return getParentItems(loc) === TOP;
    }

    function isNotTop(loc) {
        return !isTop(loc);
    }

    function getChildren(item) {
        if (!_isBranch(item)) {
            throw new Error('Tried getting children of a leaf');
        }
        return _getChildren(item);
    }

    function getLefts(loc) {
        return getPath(loc).left;
    }

    function getRights(loc) {
        return getPath(loc).right;
    }

    function hasChanged(loc) {
        return getPath(loc).changed;
    }

    function location(item, path) {
        return [item, path || {
            left: [],
            right: [],
            parentItems: TOP,
            parentPath: TOP,
            changed: false
        }];
    }

    function canGoLeft(loc) {
        return !isEmpty(getLefts(loc));
    }

    function goLeft(loc) {
        if (_isEnd(loc)) return loc;

        var item = getItem(loc);
        var path = getPath(loc);

        var leftSibling = (0, _last2.default)(path.left);
        var newLeft = (0, _init2.default)(path.left);
        var newRight = [item].concat(path.right);
        return location(leftSibling, (0, _merge2.default)(path, {
            left: newLeft,
            right: newRight
        }));
    }

    function goLeftmost(loc) {
        var path = getPath(loc);

        var hasLeft = Boolean(path.left.length);
        if (!hasLeft) return loc;

        var item = getItem(loc);
        var leftMost = (0, _head2.default)(path.left);
        var newLeft = [];
        var newRight = (0, _tail2.default)(path.left).concat([item], path.right);
        return location(leftMost, (0, _merge2.default)(path, {
            left: newLeft,
            right: newRight
        }));
    }

    function canGoRight(loc) {
        return !isEmpty(getRights(loc));
    }

    function goRight(loc) {
        var item = getItem(loc);
        var path = getPath(loc);

        var rightSibling = (0, _head2.default)(path.right);
        var newLeft = path.left.concat([item]);
        var newRight = (0, _tail2.default)(path.right);

        return location(rightSibling, (0, _merge2.default)(path, {
            left: newLeft,
            right: newRight
        }));
    }

    function goRightmost(loc) {
        var path = getPath(loc);

        var rights = getRights(loc);

        if (isEmpty(rights)) return loc;

        var lefts = getLefts(loc);
        var item = getItem(loc);
        var rightMost = (0, _last2.default)(rights);
        var newLeft = lefts.concat([item], (0, _init2.default)(rights));
        var newRight = [];
        return location(rightMost, (0, _merge2.default)(path, {
            left: newLeft,
            right: newRight
        }));
    }

    function canGoDown(loc) {
        return isBranch(loc) && !isEmpty(getChildren(getItem(loc)));
    }

    function goDown(loc) {
        if (!isBranch(loc)) return null;

        var item = getItem(loc);
        var path = getPath(loc);

        var children = getChildren(item);
        var newLeft = [];
        var newRight = (0, _tail2.default)(children);
        var newLoc = location((0, _head2.default)(children), (0, _merge2.default)(path, {
            left: newLeft,
            right: newRight,
            parentItems: (path.parentItems || []).concat([item]),
            parentPath: path
        }));
        return newLoc;
    }

    function canGoUp(loc) {
        return !isTop(loc);
    }

    function goUp(loc) {
        if (isTop(loc)) return null;

        var path = getPath(loc);
        var parent = (0, _last2.default)(path.parentItems);
        if (!hasChanged(loc)) {
            return location(parent, path.parentPath);
        }

        var parentPath = path.parentPath;
        var item = getItem(loc);
        return location(makeItem(parent, path.left.concat([item], path.right)), (0, _assoc2.default)('changed', true, parentPath));
    }

    function _canGoUpRight(loc) {
        if (!canGoUp(loc)) return false;
        var parent = goUp(loc);
        return parent ? canGoRight(parent) : false;
    }

    function _next(loc) {
        if (_isEnd(loc)) return loc;

        // $FlowIgnore: goDown(loc) cannot be null if canGoDown returns true.
        if (canGoDown(loc)) return goDown(loc);
        if (canGoRight(loc)) return goRight(loc);

        function isNotTopAndCantGoUpAndRight(currLoc) {
            return !isTop(currLoc) && !_canGoUpRight(currLoc);
        }

        var currLoc = whilst(isNotTopAndCantGoUpAndRight, goUp, loc);

        if (isTop(currLoc)) return location(getItem(currLoc), END);

        // The above predicate isNotTopAndCantGoUpAndRight makes sure
        // we can do this without returning null.
        // $FlowIgnore
        return goRight(goUp(currLoc));
    }

    function _goToRightmostChild(loc) {
        // $FlowIgnore: this will not return null (should be checked before).
        return goRightmost(goDown(loc));
    }

    function _prev(loc) {
        if (canGoLeft(loc)) {
            return whilst(isBranch, _goToRightmostChild, goLeft(loc));
        }
        return isTop(loc) ? null : goUp(loc);
    }

    function _root(loc) {
        if (_isEnd(loc)) return loc;
        return whilst(isNotTop, goUp, loc);
    }

    function _insertLeft(insertItem, loc) {
        if (isTop(loc)) {
            throw new Error('Tried inserting left of top');
        }

        var item = getItem(loc);
        var path = getPath(loc);
        return location(item, {
            left: path.left.concat([insertItem]),
            right: path.right,
            parentItems: path.parentItems,
            parentPath: path.parentPath,
            changed: true
        });
    }

    function _insertRight(insertItem, loc) {
        if (isTop(loc)) {
            throw new Error('Tried inserting left of top');
        }

        var item = getItem(loc);
        var path = getPath(loc);

        return location(item, {
            left: path.left,
            right: [insertItem].concat(path.right),
            parentItems: path.parentItems,
            parentPath: path.parentPath,
            changed: true
        });
    }

    function _remove(loc) {
        var path = getPath(loc);

        if (isTop(loc)) {
            throw new Error('Can\'t remove top.');
        }

        if (path.left.length) {
            var leftSibling = location((0, _last2.default)(path.left), (0, _merge2.default)(path, {
                left: (0, _init2.default)(path.left),
                changed: true
            }));
            return whilst(isBranch, _goToRightmostChild, leftSibling);
        } else {
            return location(makeItem((0, _last2.default)(path.parentItems), path.right), isTop(loc) ? path.parentPath : (0, _assoc2.default)('changed', true, path.parentPath));
        }
    }

    function _replace(replaceWith, loc) {
        var path = getPath(loc);
        return location(replaceWith, (0, _assoc2.default)('changed', true, path));
    }

    function _edit(fn, loc) {
        return _replace(fn(getItem(loc)), loc);
    }

    function _insertChild(item, loc) {
        var newChildren = [item].concat(getChildren(getItem(loc)));
        return _replace(makeItem(item, newChildren), loc);
    }

    function _appendChild(item, loc) {
        var newChildren = getChildren(getItem(loc)).concat([item]);
        return _replace(makeItem(item, newChildren), loc);
    }

    /**
     * A Zipper class using the implementation functions
     * provided in `makeZipper`.
     *
     * Don't use the constructor directly. Use `Zipper.from(item)`
     * for the chainable API or `Zipper.loc(item)` for the functional
     * API.
     */

    var Zipper = function () {
        _createClass(Zipper, null, [{
            key: 'from',


            /**
             * Returns a Zipper instance from item.
             */
            value: function from(item) {
                return new Zipper(location(item));
            }

            /**
             * Returns a location object from item.
             * Use the returned value if using
             * the non-chained, plain function API.
             *
             * ```javascript
             * const {
             *   down,
             *   right,
             *   loc,
             *   value,
             * } = ListZipper;
             *
             * const z = loc([1, 2, 3])
             * value(right(down(z)))
             * // 2
             * ```
             */

        }, {
            key: 'loc',
            value: function loc(item) {
                return location(item);
            }
        }]);

        function Zipper(loc) {
            _classCallCheck(this, Zipper);

            this._value = loc;
        }

        _createClass(Zipper, [{
            key: 'map',
            value: function map(fn) {
                var value = this._value;
                if (!value) return this;
                return new Zipper(fn(value));
            }

            /**
             * Calls `fn` with two arguments: the current item
             * and the current path, and returns itself.
             * @instance
             */

        }, {
            key: 'do',
            value: function _do(fn) {
                var val = this._value;
                if (!val) return this;

                fn(getItem(val), getPath(val));
                return this;
            }

            /**
             * Returns the item at this location.
             */

        }, {
            key: 'value',
            value: function value() {
                if (!this._value) return null;
                return getItem(this._value);
            }
        }, {
            key: 'path',
            value: function path() {
                return this._value ? getPath(this._value) : null;
            }

            /**
             * Moves location to the left sibling.
             */

        }, {
            key: 'left',
            value: function left() {
                return this.map(goLeft);
            }

            /**
             * Moves location to the leftmost sibling.
             */

        }, {
            key: 'leftmost',
            value: function leftmost() {
                return this.map(goLeftmost);
            }

            /**
             * Moves location to the right sibling.
             */

        }, {
            key: 'right',
            value: function right() {
                return this.map(goRight);
            }

            /**
             * Moves location to the rightmost sibling.
             */

        }, {
            key: 'rightmost',
            value: function rightmost() {
                return this.map(goRightmost);
            }

            /**
             * Moves location to the parent.
             */

        }, {
            key: 'up',
            value: function up() {
                return this.map(goUp);
            }

            /**
             * Moves location to the leftmost child.
             */

        }, {
            key: 'down',
            value: function down() {
                return this.map(goDown);
            }

            /**
             * Moves location to the root, applying
             * any changes.
             */

        }, {
            key: 'root',
            value: function root() {
                return this.map(_root);
            }

            /**
             * Moves location to the next element
             * in depth-first order..
             */

        }, {
            key: 'next',
            value: function next() {
                if (!this._value) return this;

                var newVal = _next(this._value);
                if (newVal === this._value) {
                    return this;
                }
                return new Zipper(newVal);
            }

            /**
             * Moves location to the previous element
             * in depth-first order.
             */

        }, {
            key: 'prev',
            value: function prev() {
                if (!this._value) return this;

                var newVal = _prev(this._value);
                if (newVal === this._value) {
                    return this;
                }
                return new Zipper(newVal);
            }

            /**
             * Returns true if zipper is exhausted
             * by calls to `next`.
             */

        }, {
            key: 'isEnd',
            value: function isEnd() {
                if (!this._value) return null;
                return _isEnd(this._value);
            }

            /**
             * Inserts an item as the first child.
             * Number of children grows by one.
             */

        }, {
            key: 'insertChild',
            value: function insertChild(item) {
                return this.map(_insertChild.bind(null, item));
            }

            /**
             * Inserts item as a new left sibling.
             */

        }, {
            key: 'insertLeft',
            value: function insertLeft(item) {
                return this.map(_insertLeft.bind(null, item));
            }

            /**
             * Inserts an item as a new right sibling.
             */

        }, {
            key: 'insertRight',
            value: function insertRight(item) {
                return this.map(_insertRight.bind(null, item));
            }

            /**
             * Inserts an item as the last child.
             * Number of children grows by one.
             */

        }, {
            key: 'appendChild',
            value: function appendChild(item) {
                return this.map(_appendChild.bind(null, item));
            }

            /**
             * Removes item at current location.
             */

        }, {
            key: 'remove',
            value: function remove() {
                return this.map(_remove);
            }

            /**
             * Replaces item at current location.
             */

        }, {
            key: 'replace',
            value: function replace(item) {
                return this.map(_replace.bind(null, item));
            }

            /**
             * Replaces item at current location with the return
             * value of calling `fn` with the current item.
             */

        }, {
            key: 'edit',
            value: function edit(fn) {
                return this.map(_edit.bind(null, fn));
            }
        }]);

        return Zipper;
    }();

    Zipper.up = goUp;
    Zipper.down = goDown;
    Zipper.left = goLeft;
    Zipper.leftmost = goLeftmost;
    Zipper.right = goRight;
    Zipper.rightmost = goRightmost;
    Zipper.root = _root;
    Zipper.value = getItem;
    Zipper.next = _next;
    Zipper.prev = _prev;
    Zipper.edit = _edit;
    Zipper.replace = _replace;
    Zipper.insertChild = _insertChild;
    Zipper.insertLeft = _insertLeft;
    Zipper.insertRight = _insertRight;
    Zipper.appendChild = _appendChild;
    Zipper.remove = _remove;
    Zipper.isEnd = _isEnd;


    return Zipper;
}

var ArrayZipper = makeZipper(function (arr) {
    return !!arr.length;
}, function (arr) {
    return arr;
}, function (_, children) {
    return children;
});

exports.ArrayZipper = ArrayZipper;
exports.default = makeZipper;